# library(ggplot)
library(doParallel)
library(changeS)

totalCores = detectCores()
cluster <- makeCluster(totalCores[1]-1)
registerDoParallel(cluster)

estimate_coverage_probability_fattail = function(runs = 100, n = 100){
  
  
  # Estimates the coverage probability of confidence intervals for changepoints
  # in time series data generated from fat-tailed distributions.
  #
  # Args:
  #   runs: Number of simulation replications.
  #   n: Number of points in the time series.
  #
  # Returns:
  #   A data frame containing estimated changepoints, confidence intervals, and containment information.
  
  
  ## SIMULATION:
  ## We simulate a time series of n points, each drawn from a t-distribution with 3 degrees of freedom.
  ## The mean at each point follows a linear change from mean1 to mean2 during trueChangePt1:trueChangePt2.
  ## Random noise is drawn from a fat-tailed t distribution with mean 0 and 3 degrees of freedom.
  
  

  x <- 1:n
  y <- vector(length=n)
  
  #set means
  mean1 <- 10
  mean2 <- 12.5
  trueChangePt1 <- round(n/4)
  trueChangePt2 <-round(n*2/3)
  true_changepoint = (trueChangePt1 + trueChangePt2)/2
  
  
  parallel_obj <- foreach(i = 1:runs, .combine=rbind, .packages=c("changeS", "segmented")) %dopar% {
    #function to check if true changepoint is within CI
    print(i)
    ci_contains = function(value, lower_limit, upper_limit) {
      return(value >= lower_limit & value <= upper_limit) #if the true changepoint
    }
    # data generation - simulation code goes here
    y[1:trueChangePt1] <- rt(trueChangePt1,3) + mean1 #heavy tail noise with mean1 = 10
    
    #gradual changepoint - linear path from trueChangePt1 to trueChangePt2
    y[(trueChangePt1+1):trueChangePt2] <- seq(from=mean1, to=mean2,
                                              length.out=trueChangePt2-trueChangePt1) + rt(n=trueChangePt2-trueChangePt1, 3)
    
    #heavy tail noise with mean2 = 12.5
    y[(trueChangePt2+1):n] <- rt(n-trueChangePt2,3) + mean2
    d <- data.frame(x=x,y=y)
    
    #fit S-Curve
    ours <- fitS(d, 1, 2, depth=1)
    
    
    
    #extract parameter estimates: CI boundaries and changepoint estimate
    chngpt_stderror = summary(ours$nlsOut)$coefficient[,"Std. Error"][[4]]
    chngpt = summary(ours$nlsOut)$coefficient[,"Estimate"][[4]]
    #form confidence interval
    ci_lb = chngpt - qnorm(1-.05/2)*summary(ours$nlsOut)$coefficient[,"Std. Error"][[4]]
    
    ci_ub = chngpt + qnorm(1-.05/2)*summary(ours$nlsOut)$coefficient[,"Std. Error"][[4]]
    
    #check if the true changepoint is in the confidence interval
    func_output <- ci_contains(true_changepoint,ci_lb, ci_ub)
    ci_test <- FALSE
    if (func_output == 1) {
      ci_test <- TRUE
    }
    c(chngpt, ci_lb, ci_ub, ci_test)
    
  }
  stopCluster(cluster)
  
  #extract values from parallelized object
  chngpts <- parallel_obj[,1]
  lower_bounds <- parallel_obj[,2]
  upper_bounds <- parallel_obj[,3]
  containment_tracker <- parallel_obj[,4]
  
  #return a dataframe containing all of the vectors as columns
  df = data.frame(true_changepoint = rep(true_changepoint, length(containment_tracker)),
                  changepoint = chngpts,
                  lower_bound = lower_bounds,
                  upper_bound = upper_bounds,
                  in_ci = containment_tracker)
  
  for (index in 1:nrow(df)) {
    if (df[index, "in_ci"] == 1) {
      df[index, "in_ci"] <- TRUE
    } else {
      df[index, "in_ci"] <- FALSE
    }
  }
  
  return(df)  
  
  
  
}

###
#control these hyperparameters for the simulation
replicates <- 500
n <- 100
coverage_df = estimate_coverage_probability_fattail(runs = replicates, n = n)
#convert 0s and 1s in in_ci column to bool TRUE/FALSE
coverage_df$in_ci <- ifelse(coverage_df$in_ci == 1, TRUE, FALSE)

print(paste("estimated coverage probability:", mean(coverage_df$in_ci)))


