# library(ggplot)
library(doParallel)
library(changeS)
totalCores = detectCores()
cluster <- makeCluster(totalCores[1]-1)
registerDoParallel(cluster)

estimate_coverage_probability_norm = function(runs = 100, n = 100){
  
  # Simulates time series data with a changepoint and estimates confidence intervals.
  #
  # Args:
  #   runs: Number of simulation runs.
  #   n: Number of points in the time series.
  #
  # Returns:
  #   A data frame containing estimated changepoints, confidence intervals, and containment information.
  #
  # Note:
  #   This function requires the 'changeS' and 'segmented' packages.
  
  # SIMULATION:
  # We are simulating a time series of n points, normally distributed
  # with standard deviation 2.
  # The mean at each point is mean1 for the points 1:trueChangePt1,
  # mean2 for the points trueChangePt2:n, with a linear change from
  # mean1 to mean2 during trueChangePt1:trueChangePt2. Random noise
  # is then added to the mean each point, drawn from a normal
  # distribution with mean 0 and standard deviation = 2.

  x <- 1:n
  y <- vector(length=n)
  
  #set pre changepoint mean to 10, post-changepoint mean to 12.5
  mean1 <- 10
  mean2 <- 12.5
  trueChangePt1 <- round(n/4)
  trueChangePt2 <-round(n*2/3)
  true_changepoint = (trueChangePt1 + trueChangePt2)/2
  

  parallel_obj <- foreach(i = 1:runs, .combine=rbind, .packages=c("changeS", "segmented")) %dopar% {
    #function to check if true changepoint is within CI
    ci_contains = function(value, lower_limit, upper_limit) {
      return(value >= lower_limit & value <= upper_limit) #if the true changepoint
    }
    # data generation - simulation code goes here
    y[1:trueChangePt1] <- rnorm(trueChangePt1,mean1,2) #random noise with mean 10 (mean1)
    
    #gradual changepoint along linear path
    y[(trueChangePt1+1):trueChangePt2] <- seq(from=mean1, to=mean2,
                                              length.out=trueChangePt2-trueChangePt1) + rnorm(n=trueChangePt2-trueChangePt1, sd=2)
    #random noise (normal) with mean 12.5
    y[(trueChangePt2+1):n] <- rnorm(n-trueChangePt2,mean2,2)
    d <- data.frame(x=x,y=y)
    
    #fit S-Curve
    ours <- fitS(d, 1, 2, depth=1)
    
    
    
    #extract parameter estimates: CI boundaries and changepoint estimate
    chngpt_stderror = summary(ours$nlsOut)$coefficient[,"Std. Error"][[4]]
    chngpt = summary(ours$nlsOut)$coefficient[,"Estimate"][[4]]
    #form confidence interval
    ci_lb = chngpt - qnorm(1-.05/2)*summary(ours$nlsOut)$coefficient[,"Std. Error"][[4]]
    
    ci_ub = chngpt + qnorm(1-.05/2)*summary(ours$nlsOut)$coefficient[,"Std. Error"][[4]]
    
    #check if the true changepoint is in the confidence interval
    func_output <- ci_contains(true_changepoint,ci_lb, ci_ub)
    ci_test <- FALSE
    if (func_output == 1) {
      ci_test <- TRUE
    }
    c(chngpt, ci_lb, ci_ub, ci_test)
    
  }
  stopCluster(cluster)
  
  chngpts <- parallel_obj[,1]
  lower_bounds <- parallel_obj[,2]
  upper_bounds <- parallel_obj[,3]
  containment_tracker <- parallel_obj[,4]
  
  #return a dataframe containing all of the vectors as columns
  df = data.frame(true_changepoint = rep(true_changepoint, length(containment_tracker)),
                  changepoint = chngpts,
                  lower_bound = lower_bounds,
                  upper_bound = upper_bounds,
                  in_ci = containment_tracker)
  
  for (index in 1:nrow(df)) {
    if (df[index, "in_ci"] == 1) {
      df[index, "in_ci"] <- TRUE
    } else {
      df[index, "in_ci"] <- FALSE
    }
  }
  
  return(df)  
  
  
  
}



#####
#Control These parameters 
replicates <- 500
n <- 100
coverage_df = estimate_coverage_probability_norm(runs = replicates, n = n)
coverage_df$in_ci <- ifelse(coverage_df$in_ci == 1, TRUE, FALSE)

print(paste("estimated coverage probability:", mean(coverage_df$in_ci)))



